@@ -1836,30 +1836,6 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	D_ASSERT(mdev->cstate < Connected);
 	mdev->o_state = Unknown;
 
-	/* in case we have been syncing, and then we drop the connection,
-	 * we need to "w_resume_next_sg", which we try to achieve by
-	 * setting the STOP_SYNC_TIMER bit, and schedulung the timer for
-	 * immediate execution.
-	 * unfortunately we cannot be sure that the timer already triggered.
-	 *
-	 * so we del_timer_sync here, and check that bit.
-	 * if it is still set, we queue w_resume_next_sg anyways,
-	 * just to be sure.
-	 */
-
-	del_timer_sync(&mdev->resync_timer);
-	spin_lock_irq(&mdev->req_lock);
-	if (test_and_clear_bit(STOP_SYNC_TIMER,&mdev->flags)) {
-		mdev->resync_work.cb = w_resume_next_sg;
-		if (list_empty(&mdev->resync_work.list))
-			_drbd_queue_work(&mdev->data.work,&mdev->resync_work);
-		// else: already queued, we only need to release the lock.
-	} else {
-		D_ASSERT(mdev->resync_work.cb == w_resync_inactive);
-	}
-	spin_unlock_irq(&mdev->req_lock);
-
-
 	drbd_thread_stop_nowait(&mdev->worker);
 	drbd_thread_stop(&mdev->asender);
 
@@ -2067,20 +2043,20 @@ STATIC int drbd_do_handshake(drbd_dev *mdev)
 	return 1;
 
  break_c_loop:
-	WARN( "My msock connect got accepted onto peer's sock!\n");
-	/* In case a tcp connection set-up takes longer than 
+	WARN( "Network error during initial handshake. I'll try again.\n");
+	/* In case a tcp connection set-up takes longer than
 	   connect-int, we might get into the situation that this
 	   node's msock gets connected to the peer's sock!
-	   
-	   To break out of this endless loop behaviour, we need to 
-	   wait unti the peer's msock connect tries are over. (1 Second)
 
-	   Additionally we wait connect-int/2 to hit with our next 
-	   connect try exactly in the peer's window of expectation. */
+	   To break out of this endless loop behaviour, we need to
+	   wait until the peer's msock connect tries are over. (1 Second)
+
+	   Additionally we wait connect-int/2 to increase the chance for
+	   our next connect() to hit exactly the peer's accept() window. */
 
 	set_current_state(TASK_INTERRUPTIBLE);
 	schedule_timeout(HZ + (mdev->conf.try_connect_int*HZ)/2);
-	
+
 	return 0;
 }
 
